import os
import sys
import json
import typing
import hashlib
import pathlib

import defopt

from Bio import SeqIO
from Bio.Seq import Seq

import konstel.cli as cli
import konstel.schema as schema
import konstel.formats as formats
import konstel.helpers as helpers
import konstel.encodings as encodings


def load_scheme(scheme, validate_directive=True):
    ''''''
    PACKAGE_PATH = os.path.dirname(__file__)
    scheme, _, directive = scheme.partition('.')

    # Load scheme specification
    yaml_path = pathlib.Path(f'{PACKAGE_PATH}/schemes/{scheme}/scheme.yaml')
    if not os.path.exists(yaml_path):
        raise FileNotFoundError(f'Failed to open specification for scheme {scheme}')
    spec = schema.load_scheme(yaml_path.read_text()).data

    # Validate presence and unambiguity of scheme, directive and format
    if scheme not in spec:
        raise RuntimeError(f'Unrecognised scheme {scheme}.')
    
    if validate_directive:
        if not directive:
            if len(spec[scheme]['directives']) == 1:  # One option
                directive = next(iter(spec[scheme]['directives']))
            else:
                raise RuntimeError(f"Ambiguous directive for scheme {scheme}. "
                                   f"Options: {', '.join(spec[scheme]['directives'].keys())}")
        if directive not in spec[scheme]['directives']:
            raise RuntimeError(f'Unrecognised directive {directive} for scheme {scheme}')

    return scheme, directive, spec


def prepare(string, spec):
    ''''''
    if spec['remove_whitespace']:
        string = string.translate(str.maketrans('', '', ' \n\t\r'))
    if any(spec['strip_characters']):  # any() since schema default [''] is truthy
        string = string.strip(''.join(spec['strip_characters']))
    return string


def validate_string(string, spec):
    ''''''
    if 'alphabet' in spec:
        observed = set(string)
        permitted = set(formats.alphabets[spec['alphabet']])
        if any(char not in permitted for char in observed):
            illegal_fmt = ', '.join(observed.difference(permitted))
            raise RuntimeError(f"Validation failed; illegal characters: {illegal_fmt}")
    if 'min_length' in spec:
        if len(string) < spec['min_length']:
            raise RuntimeError(f'Validation failed: min_length')
    if 'max_length' in spec:
        if len(string) > spec['max_length']:
            raise RuntimeError(f'Validation failed: max_length')
    return True


def run_directive(string, scheme, directive, spec):
    ''''''
    if 'prepare' in spec[scheme]['directives'][directive]:
        string = prepare(string, spec[scheme]['directives'][directive]['prepare'])
    if 'validate' in spec[scheme]['directives'][directive]:
        validate_string(string, spec[scheme]['directives'][directive]['validate'])
    if spec[scheme]['directives'][directive]['helper']:
        helper_name = f"{scheme}_{directive}".replace('-','_')
        helper = getattr(helpers, helper_name)
        string = helper(string)
    return string


def generate_hash(string, algorithm):
    ''''''
    hash_string = getattr(hashlib, algorithm)(string.encode()).hexdigest()
    return hash_string


def generate_encodings(hash_b16, spec, length, hide_prefix):
    ''''''
    encodings_raw = {n: getattr(encodings, m['type'])(hash_b16) for n, m in spec.items()}
    encodings_fmt = {}
    for name, encoding_raw in encodings_raw.items():
        prefix = spec[name]['prefix'] if not hide_prefix else ''
        
        # Determine length of each encoding
        if length:  # Manual override
            scheme_length = length
        elif 'length' in spec[name] and spec[name]['length']:  # In scheme and is truthy
            scheme_length = spec[name]['length']
        else:  # Fallback
            scheme_length = len(encoding_raw)

        # Handle 'hash' differently
        if name == 'hash':
            encodings_fmt[name] = f'{prefix}{encoding_raw}'
            if spec[name]['length']:
                encodings_fmt[f'{name}-{scheme_length}'] = f'{prefix}{encoding_raw[:scheme_length]}'
        else:
            encodings_fmt[name] = f'{prefix}{encoding_raw[:scheme_length]}'

        # Add separators
        if 'separator' in spec[name]:
            char = spec[name]['separator']['character']
            interval = spec[name]['separator']['interval']
            separated_chars = []
            for i in range(0, len(encodings_fmt[name])-2, interval):
                separated_chars.append(encodings_fmt[name][i+len(prefix):i+len(prefix)+interval])
            encodings_fmt[name] = prefix + char.join(separated_chars)

    return encodings_fmt


def format_encodings(outputs, output_type='dict'):
    '''Optionally prints encodings in tabular format'''
    if output_type == 'json':
        return json.dumps(outputs)  # Do nothing, return dict
    elif output_type == 'tsv':
        return '\t'.join(outputs.values())
    elif output_type == 'table':
        outputs_fmt = ''
        for k, v in outputs.items():
            outputs_fmt += f'{k:<20} {v:<15}\n'
        return outputs_fmt


def generate(scheme: str,
             file: str,
             format: typing.Union[str, None] = None,
             output: str = 'json',
             length: int = 0,
             hide_prefix: bool = False,
             sequence: str = ''):
    '''
    Generate identifier(s) for input file path or stdin according to the specified scheme
    Returns dict, and prints format specified in OUTPUT

    :arg scheme: Scheme name; use {scheme}.{directive} if scheme specifies multiple directives
    :arg file: Input path or - for stdin
    :arg format: Input format; mandatory if scheme specifies multiple formats
    :arg output: Output format for CLI (json, tsv, table)
    :arg length: Encoding length; overrides scheme
    :arg hide_prefix: Hide encoding prefix; overrides scheme
    :arg sequence: Input string, overrides file
    '''
    scheme, directive, spec = load_scheme(scheme)

    # Validate format
    if not format:
        if len(spec[scheme]['directives'][directive]['formats']) == 1:
            format = next(iter(spec[scheme]['directives'][directive]['formats']))
        else:
            raise RuntimeError(f'Ambiguous format for directive {directive} of scheme {scheme}')

    # Validate, read and format input file or stdin
    if sequence:
        string = sequence.strip()
    elif file == '-':
        string = sys.stdin.read().strip()
    elif os.path.exists(file):
        string = pathlib.Path(file).read_text().strip()
    else:
        raise FileNotFoundError(f'File {file} not found ')
    string = getattr(formats, format)(string)

    # Validate output format
    if not output in schema.OUTPUT_TYPES:
        raise RuntimeError(f'Unrecognised output type {output}. Options: {schema.OUTPUT_TYPES}')

    # Handle chained directives
    dag = [directive]
    target = spec[scheme]['directives'][directive].get('target')
    if target:
        g = {s: spec[scheme]['directives'][s].get('target') for s in spec[scheme]['directives']}
        d = target
        while d:
            dag.append(d)
            d = g[d]
    for d in dag:
        string = run_directive(string, scheme, d, spec)
    hash_b16 = generate_hash(string, spec[scheme]['algorithm'])
    encodings_ = generate_encodings(hash_b16, spec[scheme]['encodings'], length, hide_prefix)
    outputs = {**{'scheme': scheme}, **encodings_}

    return outputs


def regenerate(scheme: str,
               hash_string: str,
               output: str = 'json',
               length: int = None,
               hide_prefix: bool = False):
    '''
    Regenerate identifier(s) for an existing full length hash according to the specified scheme

    :arg scheme: Scheme name
    :arg hash_string: Full length hash digest
    :arg output: Output format (dict, tsv or table)
    :arg length: Encoding length (overrides scheme)
    :arg hide_prefix: Hide encoding prefix (overrides scheme)
    '''
    scheme, directive, spec = load_scheme(scheme, validate_directive=False)
    hash_type = spec[scheme]['encodings']['hash']['type']
    print(f"Scheme uses {hash_type}", file=sys.stderr)

    # Validate presence of 'hash' in scheme
    if not 'hash' in spec[scheme]['encodings']:
        raise RuntimeError(f'Scheme {scheme} must specify a hash encoding')

    # Validate output format
    if not output in schema.OUTPUT_TYPES:
        raise RuntimeError(f'Unrecognised output type {output}. Options: {schema.OUTPUT_TYPES}')

    # Normalise and decode supplied hash encoding
    prefix = spec[scheme]['encodings']['hash']['prefix']
    hash_string = hash_string[hash_string.startswith(prefix) and len(prefix):]  # Remove prefix
    if 'separator' in spec[scheme]['encodings']['hash']:
        sep_char = spec[scheme]['encodings']['hash']['separator']['character']
        hash_string = hash_string.replace(sep_char, '')
    hash_b16 = getattr(encodings, f'decode_{hash_type}')(hash_string)

    # Regenerate using scheme
    encodings_ = generate_encodings(hash_b16, spec[scheme]['encodings'], length, hide_prefix)
    outputs = {**{'scheme': scheme}, **encodings_}

    return outputs



import base64
import itertools
from math import ceil, log


def base32(hash_b16):
    '''Returns lowercased RFC base32 encoding of a base16 hash'''
    return base64.b32encode(bytes.fromhex(hash_b16)).decode().lower().rstrip('=')  # Remove padding


def decode_base32(hash_base32):
    '''Returns base16 encoding of case-normalised RFC base32 hash'''
    while len(hash_base32) % 8 != 0:
        hash_base32 = hash_base32 + '='  # Add padding
    return base64.b32decode(hash_base32.upper()).hex()


def cbase32(hash_b16):
    '''Returns lowercased Crockford's base32 encoding of a base16 hash'''
    hash_b32 = base32(hash_b16)
    base32_symbols = 'abcdefghijklmnopqrstuvwxyz234567'
    cbase32_symbols = '0123456789abcdefghjkmnpqrstvwxyz'
    base32_to_cbase32 = dict(zip(base32_symbols, cbase32_symbols))
    return ''.join([base32_to_cbase32[c] for c in hash_b32])


def decode_cbase32(hash_cbase32):
    '''Returns base16 from Crockford cbase32 encoded hash'''
    while len(hash_cbase32) % 8 != 0:
        hash_cbase32 = hash_cbase32 + '='  # Add padding
    base32_symbols = 'abcdefghijklmnopqrstuvwxyz234567'
    cbase32_symbols = '0123456789abcdefghjkmnpqrstvwxyz'
    cbase32_to_base32 = dict(zip(cbase32_symbols, base32_symbols))
    hash_base32 = ''.join([cbase32_to_base32.get(c, c) for c in hash_cbase32]).upper()
    return decode_base32(hash_base32)


def phonemes_10_5(hash_b16):
    '''
    Returns word comprising consonant-vowel phonemes from a base16 hash
    Maps 10 consonants and 5 vowels to base10 encoded hash
    2.8 bits per character
    '''
    hash_b10 = str(int(hash_b16, 16))
    vowel_map = dict(zip(map(str, range(10)), 'aeiou'*2))
    consonant_map = dict(zip(map(str, range(10)), 'fhklmprstv'))
    word = ''
    for i, char in enumerate(hash_b10):
        if not i % 2:
            word += consonant_map[char]
        else:
            word += vowel_map[char]
    return word


def phonemes_16_4(hash_b16, result_len=None, phonemes=None):
    '''
    Returns word comprising `hash_b16` represented by `phonemes` of max
    number `result_len`.

    This function transforms the `hash_b16` into a number of base
    `len(phonemes)` and then returns that number printed out with each digit
    represented as a phoneme. This has the advantage, relative to other phoneme
    schemes, of allowing phoneme lists of non-base-2 length.
    '''
    if phonemes is None:
        phonemes = [''.join(i) for i in itertools.product('bdfghjklmnprstvz', 'aiou')]
    hash_b2 = int(hash_b16, 16)
    if result_len is None:
        result_len = ceil(log(hash_b2, len(phonemes)))

    word = []
    for _ in range(ceil(log(hash_b2, len(phonemes)))):
        phoneme_num, hash_b2 = hash_b2 % len(phonemes), hash_b2 // len(phonemes)
        word.append(phonemes[phoneme_num])
    return ''.join(word[:-result_len-1:-1])


# def phonemes_16_4(hash_b16):  # Index-based version
#     '''
#     Returns word comprising consonant-vowel phonemes from a base16 hash
#     Maps 16 consonants and 4 vowels to six bit windows of hash
#     3 bits per character
#     '''
#     hash_b2 = bin(int(hash_b16, 16))[2:]  # Binary string from base16 hash
#     vowel_map = dict(zip(range(4), 'aiou'))
#     consonant_map = dict(zip(range(16), 'bdfghjklmnprstvz'))
#     word = ''
#     if len(hash_b2) % 6 != 0:
#         # Pad with leading zeros
#         hash_b2 = hash_b2.zfill(6 * ceil(len(hash_b2) / 6))
#     for pos in range(0, len(hash_b2), 6):
#         window = hash_b2[pos:pos+6]
#         c, v = int(window[:4], 2), int(window[4:], 2)
#         word += f'{consonant_map[c]}{vowel_map[v]}'
#     return word


# def phonemes_16_4_bits_old(hash_b16, result_len=None):
#     '''
#     Returns word comprising consonant-vowel phonemes from a base16 hash
#     Maps 16 consonants and 4 vowels to six bit windows of hash
#     3 bits per character

#     Uses bit-wise logic to calculate which phonemes are which.
#     '''
#     phonemes = [''.join(i) for i in itertools.product('bdfghjklmnprstvz', 'aiou')]
#     phoneme_bit_size = log2(len(phonemes))
#     assert phoneme_bit_size.is_integer(), 'Must have a power of two number of phonemes'

#     hash_b2 = int(hash_b16.hexdigest(), 16)

#     if result_len is None:
#         result_len = int(8 * hash_b16.digest_size / phoneme_bit_size)
#     front_offset = int(phoneme_bit_size * result_len - phoneme_bit_size)
#     mask = (2 ** int(phoneme_bit_size) - 1) << front_offset

#     word = ''
#     for _ in range(result_len):
#         phoneme_num = (hash_b2 & mask) >> front_offset
#         word += phonemes[phoneme_num]
#         hash_b2 = hash_b2 << int(phoneme_bit_size)
#     return word


import io

from Bio import SeqIO


alphabets = {
	'nucleotide': 'ACGTU',
	'nucleotide-ambiguous': '-ABCDGHKMNRSTUVWY',
	'protein': '*ACDEFGHIKLMNPQRSTVWY'
}


def string(string):
	'''
	Generic string format
	'''
	return str(string)

def fasta(string):
	'''
	Returns uppercased sequence from fasta string
	Prepends '>' if absent to enable naked sequence parsing
	'''
	if '>' not in string:
		string = f'>\n{string}'
	string_as_file = io.StringIO(string)
	record = SeqIO.read(string_as_file, 'fasta')
	assert len(record.seq) >= 1, 'Sequence is empty'
	return str(record.seq).upper()


import hashlib
import inspect

from strictyaml import load, Bool, Int, Str, Seq, Map, Enum, MapPattern, Optional, YAMLValidationError

from konstel import formats
from konstel import encodings


ALGORITHMS = hashlib.algorithms_available
ALPHABETS = formats.alphabets.keys()
BASE_ENCODINGS = {'base32', 'cbase32'}
ENCODINGS = {o[0] for o in inspect.getmembers(encodings, inspect.isfunction)}
FORMATS = {o[0] for o in inspect.getmembers(formats, inspect.isfunction)}
OUTPUT_TYPES = {'json', 'tsv', 'table'}


def load_scheme(yaml_text):
    '''
    Some optional keys have enforced default values, otherwise use dict.get()
    '''
    schema = MapPattern(
        Str(), Map({
            'description': Str(),
            Optional('alias'): Str(),
            'version': Str(),
            'directives': MapPattern(
                Str(), Map({
                    Optional('description'): Str(),
                    'formats': Seq(Enum(FORMATS)),
                    Optional('prepare'): Map({
                        Optional('remove_whitespace', default=False): Bool(),
                        Optional('strip_characters', default=['']): Seq(Str()),
                    }),
                    Optional('validate'): Map({
                        Optional('alphabet'): Enum(ALPHABETS),
                        Optional('min_length'): Int(),
                        Optional('max_length'): Int(),
                    }),
                    Optional('target'): Str(),
                    Optional('helper', default=False): Bool()
                }),
            ),
            'algorithm': Enum(ALGORITHMS),
            'encodings': MapPattern(
                Str(), Map({
                    'type': Enum(ENCODINGS),
                    Optional('length', default=0): Int(),
                    Optional('prefix', default=''): Str(),
                    Optional('separator'): Map({
                        'character': Str(),
                        'interval': Int()
                    })
                })
            )
        })
    )
    return load(yaml_text, schema)



import typing

import defopt

from konstel import konstel


def generate(scheme: str,
             file: str,
             format: typing.Union[str, None] = None,
             output: str = 'json',
             length: int = 0,
             hide_prefix: bool = False):
    '''
    Generate identifier(s) for input file path or stdin according to the specified scheme
    Returns dict, and prints format specified in OUTPUT

    :arg scheme: Scheme name; use {scheme}.{directive} if scheme specifies multiple directives
    :arg file: Input path or - for stdin
    :arg format: Input format; mandatory if scheme specifies multiple formats
    :arg output: Output format (json, tsv, table)
    :arg length: Encoding length; overrides scheme
    :arg hide_prefix: Hide encoding prefix; overrides scheme
    '''
    outputs = konstel.generate(scheme, file, format, output, length, hide_prefix)
    print(konstel.format_encodings(outputs, output))


def regenerate(scheme: str,
               hash_string: str,
               output: str = 'json',
               length: int = None,
               hide_prefix: bool = False):
    '''
    Regenerate identifier(s) for an existing full length hash according to the specified scheme

    :arg scheme: Scheme name
    :arg hash_string: Full length hash digest
    :arg output: Output format (dict, tsv or table)
    :arg length: Encoding length (overrides scheme)
    :arg hide_prefix: Hide encoding prefix (overrides scheme)
    '''
    outputs = konstel.regenerate(scheme, hash_string, output, length, hide_prefix)
    print(konstel.format_encodings(outputs, output))


def main():
    defopt.run(
        {'gen': generate, 'regen': regenerate},
        strict_kwonly=False, no_negated_flags=True)


import konstel

from setuptools import setup


with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
    name = 'konstel',
    version = konstel.__version__,
    description = 'Hash-based phonemic sequence identifiers',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url = 'http://github.com/bede/konstel',
    author = 'Bede Constantinides',
    author_email = 'bedeabc@gmail.com',
    license = 'LICENSE',
    packages=['konstel'],
    include_package_data=True,  # MANIFEST.in
    python_requires='>=3.6',
    install_requires=[
        'defopt>=6.1.0',
        'strictyaml>=1.4.0',
        'biopython>=1.78',
        'parasail==1.2.4'],
    entry_points = {
        'console_scripts':['konstel=konstel.cli:main']},
    classifiers=[
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English'])
